from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

from sklik.api import SklikApi
from sklik.object import Account
from sklik.util import SKLIK_DATE_FORMAT


def _restriction_filter_validator(
        since: Optional[str],
        until: Optional[str],
        restriction_filter: Optional[Dict]) -> Dict[str, Any]:
    since = since if since else (datetime.today() - timedelta(days=29)).strftime(SKLIK_DATE_FORMAT)
    until = until if until else (datetime.today() - timedelta(days=1)).strftime(SKLIK_DATE_FORMAT)

    restriction_filter = restriction_filter or {}
    restriction_filter.setdefault("dateFrom", since)
    restriction_filter.setdefault("dateTo", until)

    return restriction_filter


def _display_option_validator(granularity: Optional[str]) -> Dict[str, Any]:
    granularity = granularity if granularity else "daily"
    return {"statGranularity": granularity}


class Report:
    """
    Iterator-based interface for handling Sklik API reports.

    Provides functionality to access and iterate through paginated report data
    from the Sklik advertising platform. The class handles pagination automatically        and formats the data into a consistent structure.

    Attributes:
        account_id: Identifier of the Sklik account.
        report_id: Unique identifier of the report.
        service: Name of the Sklik service providing the report.
        fields: List of field names to be included in the report data.

    Iterator Behavior:
        The class implements the iterator protocol, allowing for easy iteration
        over all report items. Pagination is handled automatically during iteration.

    Example:
        >>> report = Report(
        ...     account_id=123456,
        ...     report_id="abc123",
        ...     service="stats",
        ...     fields=["impressions", "clicks", "cost"]
        ... )
        >>>
        >>> # Iterate through all items
        >>> for item in report:
        ...     print(item["impressions"], item["clicks"])
        >>>
        >>> # Load specific page
        >>> first_page = report.load_page(offset=0, limit=100)
    """

    def __init__(
            self,
            account_id: int,
            report_id: str,
            service: str,
            fields: List[str],
            total_count: int = 0,
            api: Optional[SklikApi] = None) -> None:
        self.account_id = account_id
        self.report_id = report_id
        self.fields = fields
        self.service = service

        self._total_count = total_count
        self._api = api or SklikApi.get_default_api()

        self._current_offset = 0
        self._current_limit = 100
        self._current_page = self.load_page(self._current_offset, self._current_limit)
        self._current_index = 0

    def __repr__(self) -> str:
        return f"Report(account_id={self.account_id}, report_id={self.report_id}, total_count={self._total_count})"

    def __len__(self) -> int:
        return self._total_count

    def __iter__(self) -> "Report":
        return self

    def __next__(self) -> Dict[str, Any]:
        if self._current_index >= len(self._current_page):
            self._current_offset += self._current_limit
            self._current_page = self.load_page(self._current_offset, self._current_limit)
            self._current_index = 0

        if not self._current_page or self._current_index >= len(self._current_page):
            raise StopIteration()

        item = self._current_page[self._current_index]
        self._current_index += 1
        return item

    def _load_raw_page(self, offset: int, limit: int) -> List[Dict[str, Any]]:
        """
        Load and format a specific page of report data.

        Retrieves raw data from the API and transforms it into a flat structure
        where each dictionary represents one time unit with its associated metrics.

        Args:
            offset: Starting position in the complete dataset.
            limit: Maximum number of items to retrieve.

        Returns:
            List of dictionaries, each containing the requested fields
            and their values for a specific time unit.
        """

        args = [
            {"userId": self.account_id},
            self.report_id,
            {"offset": offset, "limit": limit, "displayColumns": self.fields}
        ]

        return self._api.call(self.service, "readReport", args=args)["report"]

    def load_page(self, offset: int, limit: int) -> List[Dict[str, Any]]:
        """
        Retrieve raw report data from the Sklik API.

        Makes the actual API call to fetch report data with the specified
        pagination parameters.

        Args:
            offset: Starting position for data retrieval.
            limit: Maximum number of items to retrieve.

        Returns:
            Raw API response containing the requested report data.
        """

        report = self._load_raw_page(offset, limit)
        output = [dict(**stat,
                       **{key: value for key, value in item.items() if key != "stats"})
                  for item in report for stat in item.get("stats", [])]

        return output


def create_report(
        account: Union[Account, str, int],
        service: str,
        fields: list[str],
        since: Optional[str] = None,
        until: Optional[str] = None,
        granularity: Optional[str] = None,
        restriction_filter: Optional[str] = None) -> Report:
    """
    Create a new report for the specified Sklik service.

    Creates and initializes a new report based on the provided parameters,
    allowing for flexible data retrieval and filtering options.

    Args:
        account: Account identifier. Can be an Account object, account ID string,
            or integer account number.
        service: Name of the Sklik service to generate the report for
            (e.g., 'campaigns', 'stats', 'ads').
        fields: List of field names to include in the report.
        since: Start date for the report data in 'YYYY-MM-DD' format.
            If not specified, no start date filter is applied.
        until: End date for the report data in 'YYYY-MM-DD' format.
            If not specified, no end date filter is applied.
        granularity: Time granularity for the report data
            (e.g., 'daily', 'weekly', 'monthly').
        restriction_filter: Additional filtering criteria for the report data.

    Returns:
        A Report object containing the initialized report data and metadata.

    Example:
        >>> report = create_report(
        ...     account="123456",
        ...     service="campaigns",
        ...     fields=["clicks", "impressions"],
        ...     since="2024-01-01",
        ...     until="2024-01-31",
        ...     granularity="daily"
        ... )
    """

    args = [
        _restriction_filter_validator(since, until, restriction_filter),
        _display_option_validator(granularity)
    ]

    response = account.call(service, "createReport", args)
    return Report(account.account_id, response["reportId"], service=service, fields=fields,
                  total_count=response["totalCount"], api=account.api)
